module Data.Whole where

import "base" Prelude (Num (negate, signum))
import qualified "base" Prelude as Base
import "base" Control.Arrow ((***))
import qualified "base" Data.Semigroup as Base

import Data.Difference (Difference (..))
import qualified Data.Difference as D

data Np = X1 Np
        | X0 Np
        | One
  deriving (Base.Eq, Base.Show)

instance PartialEq Np where (≡) = (Base.==)
instance Preord Np where (≤) = (Base.<=)
instance Eq Np
instance PartialOrd Np where tryCompare x y = Just (compare x y)
instance Ord Np where compare = Base.compare

instance {-# OVERLAPPING #-} Semigroup (Sum Np) where
    Sum x <> Sum y = Sum (add x y)

instance {-# OVERLAPPING #-} Semigroup (Product Np) where
    Product x <> Product y = Product (times x y)

instance {-# OVERLAPPING #-} Monoid (Product Np) where
    mempty = Product One

add :: Np -> Np -> Np
add = go False where
    go c = curry $ \ case
        (X1 x, X1 y) -> bool X0 X1 c (go True  x y)
        (X1 x, X0 y) -> bool X1 X0 c (go c     x y)
        (X1 x, One)   -> bool X0 X1 c (plus1 x)
        (X0 x, X0 y) -> bool X0 X1 c (go False x y)
        (X0 x, One)
          | not c    -> X1 x
          | True     -> X0 (plus1 x)
        (One,   One)   -> bool X0 X1 c One
        (x,    y)    -> go c y x

doubleMinus1 :: Np -> Np
doubleMinus1 = \ case
    X1 x -> X1 (X0 x)
    X0 x -> X1 (doubleMinus1 x)
    One -> One

instance Base.Ord Np where
    compare = go EQ where
        go r = curry $ \ case
            (One, One) -> r
            (One, _)  -> LT
            (_,  One) -> GT
            (X1 x, X1 y) -> go r  x y
            (X1 x, X0 y) -> go GT x y
            (X0 x, X1 y) -> go LT x y
            (X0 x, X0 y) -> go r  x y

times :: Np -> Np -> Np
times x y = case x of
    X1 x -> X0 (times x y) + y
    X0 x -> X0 (times x y)
    One -> y

plus1 :: Np -> Np
plus1 = \ case
    X1 x -> X0 (plus1 x)
    X0 x -> X1 x
    One -> X0 One

toNaturalNp :: Np -> Natural
toNaturalNp = getSum . flip stimes (Sum 1)

stimes :: Semigroup a => Np -> a -> a
stimes One = id
stimes (X0 x) = stimes x & join (<>)
stimes (X1 x) = \ a -> a <> stimes (X0 x) a

subNp' :: Np -> Np -> Difference Np
subNp' (X1 x) (X1 y) = X0 <$> subNp' x y
subNp' (X1 x) (X0 y) = case subNp' x y of
    Neg z -> Neg (doubleMinus1 z)
    Zero   -> Pos One
    Pos z -> Pos (X1 z)
subNp' (X0 x) (X1 y) = case subNp' x y of
    Neg z -> Neg (X1 z)
    Zero   -> Neg One
    Pos z -> Pos (doubleMinus1 z)
subNp' (X0 x) (X0 y) = X0 <$> subNp' x y
subNp' x One = maybe Zero Pos $ minus1 x
subNp' One y = maybe Zero Neg $ minus1 y

minus1 :: Np -> Maybe Np
minus1 One = Nothing
minus1 (X1 n) = Just (X0 n)
minus1 (X0 One) = Just One
minus1 (X0 n) = X1 <$> minus1 n

data N = Nul
       | Nonnul Np
  deriving (Base.Eq, Base.Ord, Base.Show)

instance PartialEq N where (≡) = (Base.==)
instance Preord N where (≤) = (Base.<=)
instance Eq N
instance PartialOrd N where tryCompare x y = Just (compare x y)
instance Ord N where compare = Base.compare

instance {-# OVERLAPPING #-} Semigroup (Sum N) where
    Sum Nul <> y = y
    x <> Sum Nul = x
    Sum (Nonnul x) <> Sum (Nonnul y) = Nonnul <$> Sum x <> Sum y

instance {-# OVERLAPPING #-} Monoid (Sum N) where
    mempty = Sum Nul

instance {-# OVERLAPPING #-} Semigroup (Product N) where
    Product Nul <> _ = Product Nul
    _ <> Product Nul = Product Nul
    Product (Nonnul x) <> Product (Nonnul y) = (Product . Nonnul) (x * y)

instance {-# OVERLAPPING #-} Monoid (Product N) where
    mempty = Product (Nonnul One)

instance FromInteger N where
    fromInteger 0 = Nul
    fromInteger n = Nonnul One + fromInteger (n-1)

subN' :: N -> N -> Maybe N
subN' a Nul = Just a
subN' Nul _ = Nothing
subN' (Nonnul a) (Nonnul b) = case subNp' a b of
    Pos c -> Just (Nonnul c)
    Zero -> Just Nul
    _ -> Nothing

fromNaturalN :: Natural -> N
fromNaturalN = maybe Nul (Nonnul . getSum) . flip Base.stimes (Just (Sum One))

mtimes :: Monoid a => N -> a -> a
mtimes Nul = pure mempty
mtimes (Nonnul x) = stimes x

toNaturalN :: N -> Natural
toNaturalN Nul = 0
toNaturalN (Nonnul x) = toNaturalNp x

isTopMoreInformative :: N -> N -> N -> N -> Maybe (N, N)
isTopMoreInformative a b c d = [(x, y) | x <- subN' a c, y <- subN' b d, any (> Nul) [x, y]]

isQuadraticTopMoreInformative :: N -> N -> N -> N -> N -> N -> N -> N -> Maybe (N, N, N, N)
isQuadraticTopMoreInformative a b c d e f g h =
    [(w, x, y, z) | w <- subN' a e, x <- subN' b f, y <- subN' c g, z <- subN' d h
                  , any (> Nul) [w, x, y, z]]

type Z = Difference Np

instance Base.Num Z where
    Zero + y = y
    x + Zero = x
    Pos x + Pos y = Pos (x + y)
    Pos x + Neg y = subNp' x y
    Neg x + Pos y = subNp' y x
    Neg x + Neg y = Neg (x + y)
    (*) = liftA2 (*)
    abs = maybe Zero Pos . altMap Just
    signum = (One <$)
    negate = D.negate
    fromInteger = fromIntegerZ

instance Base.Real Z where
    toRational = Base.toRational . toIntegerZ

instance Base.Enum Z where
    enumFrom = Base.iterate (+1)
    enumFromTo a b = Base.takeWhile (≤ b) (Base.enumFrom a)
    enumFromThen a a' = Base.iterate (+ (a' - a)) a
    enumFromThenTo a a' b = Base.takeWhile (≤ b) (Base.enumFromThen a a')
    toEnum = Base.fromInteger . Base.toEnum
    fromEnum = Base.fromEnum . toIntegerZ

instance Base.Integral Z where
    toInteger = toIntegerZ
    quotRem = (fromIntegerZ *** fromIntegerZ) ∘∘ (Base.quotRem `on` toIntegerZ)
    divMod = (fromIntegerZ *** fromIntegerZ) ∘∘ (Base.divMod `on` toIntegerZ)

instance FromInteger Z where fromInteger = fromIntegerZ

fromIntegerZ :: Base.Integral a => a -> Z
fromIntegerZ = \ x -> case Base.compare x 0 of
    EQ -> Zero
    GT -> Pos (toNp x)
    LT -> Neg (toNp (negate x))
  where
    toNp :: Base.Integral a => a -> Np
    toNp 1 = One
    toNp x = bool X0 X1 (0 Base./= Base.mod x 2) (toNp (Base.div x 2))

toIntegerZ :: Z -> Base.Integer
toIntegerZ = \ case
    Zero -> 0
    Pos n -> id     . Base.fromIntegral $ toNaturalNp n
    Neg n -> negate . Base.fromIntegral $ toNaturalNp n

gtimes :: Group a => Z -> a -> a
gtimes Zero = pure mempty
gtimes (Pos x) = stimes x
gtimes (Neg x) = stimes x . invert

outside :: Foldable f => f Z -> Z
outside = getSum . foldMap (Sum . signum)

abs :: Z -> N
abs = maybe Nul Nonnul . altMap Just

fromNZ :: N -> Z
fromNZ Nul = Zero
fromNZ (Nonnul x) = Pos x
